version_info = (4, 13, 1)
__version__ = ".".join([str(v) for v in version_info])


"""
4.13.0
- fix tests
- fix cov
- fix docs
- some api additions

4.12.1

- support up to py3.13

4.12.0

- remove resource module
- migrate resource module to shopyo_theme

4.11.1

- explicit errors on missing imports while looking for create_app

4.11.0

- auth is now shopyo_auth
- dashboard is now shopyo_dashboard
- app admin is now shopyo_appadmin

4.10.0

- base module now ships in a separate package: shopyo_base

4.9.5

- initialize supports sqlalchemy 2.0

4.9.4

- loosen sqlalchemy requirment; initialize still expected to fail, but package usable

4.9.3

- Remove marshmallow from deps

4.9.2

- Fix missing import

4.9.1

- Important security fixes to box__default

4.9.0

- Provide more info on missing installed_packages
- Bound SQLAlchemy version

4.8.6

- Fixes for sqlalchemy db & other upgrade supports

4.8.5

- Tests and docs fixes

4.8.4

- Implement debug mode
- Fix docs builds fail

4.8.3

- Fix load erros instead of passing

4.8.2

- Fix admin required on dashbaord

4.8.1

- Fix import errors

4.8.0

- box__bizhelp removed
- i18n and page moved to box__default

4.7.0

- i18n package
- page module revamp

4.6.0

- Packages added!!!

4.5.9

- fix autoload for packages

4.5.8

- fix key error in cmd_helper

4.5.7

- fix verbose=vebose

4.5.6

- rm app.txt for new projects

4.5.5

- Fix empty static folder for projects not using -m

4.5.4

- Refactor move register_devstatic to api

4.5.3

- Refactor load_extensions to init.py

4.5.2

- Fix static path

4.5.1

- Fix path parameter in api.assets.get_static

4.5.0

- Add url_prefix conflict detection to audit command

4.4.3

- Fix initialise, add SHOPYO_CONFIG_PROFILE env var to denote configuration profiles

4.4.2

- Fix email-validator version

4.4.1

- Fix app.py on new projects
- Fix validator required in reqs
- Fix remove settings functions from www.view

4.4.0

- Add global configs via module

4.3.7

- Remove 3.6 support

"""



"""
response.py

Utilities for standardized API responses in Flask.
"""

from flask import jsonify


def json_response(data=None, status=200, message=None, **kwargs):
    """
    Return a standard JSON response.

    Args:
        data (dict or list, optional): The main data payload.
        status (int): HTTP status code.
        message (str, optional): Optional message.
        **kwargs: Additional fields to include.

    Returns:
        Flask Response: JSON response.
    """
    response = {
        "success": 200 <= status < 300,
        "data": data,
        "message": message,
    }
    response.update(kwargs)
    return jsonify(response), status


def error_response(message, status=400, errors=None, **kwargs):
    """
    Return a standard error JSON response.

    Args:
        message (str): Error message.
        status (int): HTTP status code.
        errors (dict or list, optional): Additional error details.
        **kwargs: Additional fields to include.

    Returns:
        Flask Response: JSON response.
    """
    response = {
        "success": False,
        "error": {
            "message": message,
            "details": errors,
        },
    }
    response.update(kwargs)
    return jsonify(response), status


def paginated_response(
    items, total, page, per_page, status=200, message=None, **kwargs
):
    """
    Return a paginated JSON response.

    Args:
        items (list): List of items for the current page.
        total (int): Total number of items.
        page (int): Current page number.
        per_page (int): Items per page.
        status (int): HTTP status code.
        message (str, optional): Optional message.
        **kwargs: Additional fields to include.

    Returns:
        Flask Response: JSON response.
    """
    response = {
        "success": 200 <= status < 300,
        "data": items,
        "pagination": {
            "total": total,
            "page": page,
            "per_page": per_page,
            "pages": (total + per_page - 1) // per_page,
        },
        "message": message,
    }
    response.update(kwargs)
    return jsonify(response), status



import pytest
from flask import Flask
from shopyo.api.templates import yo_render, yo_render_string, yo_get_macro, yo_safe


def test_yo_render():
    app = Flask(__name__)
    with app.test_request_context():
        # Just check it calls render_template which will fail because template doesn't exist
        # but this confirms the function is called and reaches the Flask call
        try:
            yo_render("non_existent.html", {"x": 1})
        except Exception:
            pass


def test_yo_render_string():
    app = Flask(__name__)
    with app.test_request_context():
        result = yo_render_string("Hello {{ name }}!", {"name": "World"})
        assert result == "Hello World!"


def test_yo_safe():
    result = yo_safe("<b>bold</b>")
    assert str(result) == "<b>bold</b>"
    try:
        from markupsafe import Markup
    except ImportError:
        from flask import Markup
    assert isinstance(result, Markup)



import pytest
from flask import Flask
from shopyo.api.assets import get_static, register_devstatic


def test_get_static_debug_true():
    app = Flask(__name__)
    app.config["DEBUG"] = True
    app.config["SERVER_NAME"] = "localhost"
    register_devstatic(app, "/tmp/modules")
    with app.app_context():
        with app.test_request_context():
            url = get_static("box__default/auth", "style.css")
            assert "/devstatic/box__default/auth/f/style.css" in url


def test_get_static_debug_false():
    app = Flask(__name__)
    app.config["DEBUG"] = False
    app.config["SERVER_NAME"] = "localhost"
    with app.app_context():
        with app.test_request_context():
            url = get_static("box__default/auth", "style.css")
            assert "/static/modules/box__default/auth/style.css" in url



from unittest.mock import MagicMock, patch
from shopyo.api.forms import flash_errors


def test_flash_errors():
    mock_form = MagicMock()
    mock_form.errors = {"field1": ["error1", "error2"]}
    mock_form.field1.label.text = "Field 1"

    with patch("shopyo.api.forms.flash") as mock_flash:
        with patch(
            "shopyo.api.forms.notify_warning", side_effect=lambda x: x
        ) as mock_notify:
            flash_errors(mock_form)
            assert mock_flash.call_count == 2
            mock_flash.assert_any_call("Error in the Field 1 field - error1")
            mock_flash.assert_any_call("Error in the Field 1 field - error2")



from unittest.mock import patch, MagicMock
import os
import sys
from shopyo.api.database import autoload_models


def test_autoload_models():
    # Mock os.listdir to simulate a modules directory
    def listdir_side_effect(path):
        if path == "modules":
            return ["box__default", "test_module"]
        if path == os.path.join("modules", "box__default"):
            return ["auth"]
        return []

    # Mock init module
    mock_init = MagicMock()
    mock_init.installed_packages = ["plugin1"]

    with patch.dict("sys.modules", {"init": mock_init}):
        with patch("shopyo.api.database.os.listdir", side_effect=listdir_side_effect):
            # Patch import_module in the database module's namespace
            with patch("shopyo.api.database.importlib.import_module") as mock_import:
                with patch("shopyo.api.database.click.echo"):
                    autoload_models(verbose=True)

                    # Check some expected imports
                    mock_import.assert_any_call("modules.box__default.auth.models")
                    mock_import.assert_any_call("modules.test_module.models")
                    mock_import.assert_any_call("plugin1.models")



"""
endpoint.py

A powerful utility for building RESTful API endpoints in Flask applications.
Provides decorators and base classes for rapid API development with common patterns.
"""

from functools import wraps
from typing import Any, Callable, Dict, List, Optional, Type, TypeVar, Union
from flask import request, current_app
from sqlalchemy.orm import Query

from .models import YoModel
from .response import json_response, error_response, paginated_response

T = TypeVar("T", bound=YoModel)


class APIEndpoint:
    """
    A utility class for building RESTful API endpoints with common patterns.
    Provides decorators and methods for handling CRUD operations, validation,
    and response formatting.
    """

    def __init__(self, model_class: Type[T]):
        """
        Initialize the API endpoint with a model class.

        Args:
            model_class (Type[T]): The SQLAlchemy model class to use
        """
        self.model_class = model_class

    def validate_request(self, data: Dict[str, Any]) -> List[str]:
        """
        Validate request data. Override this method in subclasses.

        Args:
            data (Dict[str, Any]): Request data to validate

        Returns:
            List[str]: List of validation errors, empty if valid
        """
        return []

    def format_response(self, data: Any) -> Dict[str, Any]:
        """
        Format response data. Override this method in subclasses.

        Args:
            data (Any): Data to format

        Returns:
            Dict[str, Any]: Formatted response data
        """
        return data

    def get_query_filters(self) -> Dict[str, Any]:
        """
        Get query filters from request args. Override this method in subclasses.

        Returns:
            Dict[str, Any]: Query filters
        """
        return {}

    def handle_list(self, page: int = 1, per_page: int = 20) -> tuple:
        """
        Handle GET request for listing items with pagination.

        Args:
            page (int): Page number
            per_page (int): Items per page

        Returns:
            tuple: Flask response tuple
        """
        try:
            filters = self.get_query_filters()
            query = self.model_class.query.filter_by(**filters)

            # Handle search if search parameter is present
            if "search" in request.args:
                search_query = request.args.get("search", "")
                if hasattr(self.model_class, "search"):
                    query = self.model_class.search(
                        search_query, ["name", "description"]
                    )

            # Handle sorting
            if "sort_by" in request.args:
                sort_by = request.args.get("sort_by")
                if hasattr(self.model_class, sort_by):
                    sort_field = getattr(self.model_class, sort_by)
                    order = request.args.get("order", "asc")
                    if order == "desc":
                        query = query.order_by(sort_field.desc())
                    else:
                        query = query.order_by(sort_field.asc())

            pagination = query.paginate(page=page, per_page=per_page)
            items = [self.format_response(item) for item in pagination.items]

            return paginated_response(
                items=items, total=pagination.total, page=page, per_page=per_page
            )
        except Exception as e:
            current_app.logger.error(f"Error in list endpoint: {str(e)}")
            return error_response("Failed to fetch items", status=500)

    def handle_create(self) -> tuple:
        """
        Handle POST request for creating a new item.

        Returns:
            tuple: Flask response tuple
        """
        try:
            data = request.get_json()
            if not data:
                return error_response("No data provided", status=400)

            # Validate request data
            errors = self.validate_request(data)
            if errors:
                return error_response("Validation failed", errors=errors, status=400)

            # Create new instance
            instance = self.model_class.create(**data)
            return json_response(
                data=self.format_response(instance),
                message="Item created successfully",
                status=201,
            )
        except Exception as e:
            current_app.logger.error(f"Error in create endpoint: {str(e)}")
            return error_response("Failed to create item", status=500)

    def handle_get(self, item_id: Union[int, str]) -> tuple:
        """
        Handle GET request for retrieving a single item.

        Args:
            item_id (Union[int, str]): Item ID

        Returns:
            tuple: Flask response tuple
        """
        try:
            instance = self.model_class.get_by_id(item_id)
            if not instance:
                return error_response("Item not found", status=404)

            return json_response(data=self.format_response(instance))
        except Exception as e:
            current_app.logger.error(f"Error in get endpoint: {str(e)}")
            return error_response("Failed to fetch item", status=500)

    def handle_update(self, item_id: Union[int, str]) -> tuple:
        """
        Handle PUT/PATCH request for updating an item.

        Args:
            item_id (Union[int, str]): Item ID

        Returns:
            tuple: Flask response tuple
        """
        try:
            instance = self.model_class.get_by_id(item_id)
            if not instance:
                return error_response("Item not found", status=404)

            data = request.get_json()
            if not data:
                return error_response("No data provided", status=400)

            # Validate request data
            errors = self.validate_request(data)
            if errors:
                return error_response("Validation failed", errors=errors, status=400)

            # Update instance
            instance.update(**data)
            return json_response(
                data=self.format_response(instance), message="Item updated successfully"
            )
        except Exception as e:
            current_app.logger.error(f"Error in update endpoint: {str(e)}")
            return error_response("Failed to update item", status=500)

    def handle_delete(self, item_id: Union[int, str]) -> tuple:
        """
        Handle DELETE request for removing an item.

        Args:
            item_id (Union[int, str]): Item ID

        Returns:
            tuple: Flask response tuple
        """
        try:
            instance = self.model_class.get_by_id(item_id)
            if not instance:
                return error_response("Item not found", status=404)

            instance.delete()
            return json_response(message="Item deleted successfully")
        except Exception as e:
            current_app.logger.error(f"Error in delete endpoint: {str(e)}")
            return error_response("Failed to delete item", status=500)


def api_endpoint(model_class: Type[T]):
    """
    Decorator to create RESTful API endpoints for a model.

    Args:
        model_class (Type[T]): The SQLAlchemy model class to use

    Returns:
        Callable: Decorated function that returns APIEndpoint instance
    """

    def decorator(f: Callable) -> Callable:
        @wraps(f)
        def wrapper(*args, **kwargs):
            endpoint = APIEndpoint(model_class)
            return f(endpoint, *args, **kwargs)

        return wrapper

    return decorator



import os
import sys
import click
import pytest
from click.testing import CliRunner
from unittest.mock import patch, MagicMock

# Mock app module before importing cli
sys.modules["app"] = MagicMock()
sys.modules["app"].create_app = MagicMock(return_value=MagicMock())
sys.modules["shopyo.app"] = MagicMock()
sys.modules["shopyo.app"].create_app = sys.modules["app"].create_app

from shopyo.api.cli import cli


@pytest.fixture
def runner():
    return CliRunner()


def test_testok(runner):
    result = runner.invoke(cli, ["testok"])
    assert result.exit_code == 0
    assert "test ok!" in result.output


@patch("shopyo.api.cli._create_box")
def test_startbox(mock_create_box, runner):
    original_exists = os.path.exists

    def exists_side_effect(path):
        if "box__test" in path:
            return False
        return original_exists(path)

    with patch("os.path.exists", side_effect=exists_side_effect):
        result = runner.invoke(cli, ["startbox", "box__test"])
        assert result.exit_code == 0
        mock_create_box.assert_called_once_with("box__test", verbose=False)


@patch("shopyo.api.cli._create_box")
def test_startbox_exists(mock_create_box, runner):
    with patch("os.path.exists", return_value=True):
        result = runner.invoke(cli, ["startbox", "box__test"])
        assert result.exit_code == 1
        assert "already exists" in result.output


@patch("shopyo.api.cli._create_module")
@patch("shopyo.api.cli.get_module_path_if_exists")
def test_startapp_simple(mock_get_path, mock_create_module, runner):
    mock_get_path.return_value = None
    result = runner.invoke(cli, ["startapp", "testmod"])
    assert result.exit_code == 0
    mock_create_module.assert_called_once()
    args, kwargs = mock_create_module.call_args
    assert args[0] == "testmod"
    assert kwargs["base_path"] == os.path.join("modules", "", "testmod")


@patch("shopyo.api.cli._create_box")
@patch("shopyo.api.cli._create_module")
@patch("shopyo.api.cli.get_module_path_if_exists")
def test_startapp_in_box(mock_get_path, mock_create_module, mock_create_box, runner):
    mock_get_path.return_value = None
    result = runner.invoke(cli, ["startapp", "testmod", "box__test"])
    assert result.exit_code == 0
    mock_create_module.assert_called_once()
    args, kwargs = mock_create_module.call_args
    assert args[0] == "testmod"
    assert kwargs["base_path"] == os.path.join("modules", "box__test", "testmod")


def test_startapp_invalid_box_name(runner):
    result = runner.invoke(cli, ["startapp", "testmod", "invalidbox"])
    assert result.exit_code == 1
    assert "BOXNAME should start with 'box__'" in result.output


def test_startapp_invalid_mod_name(runner):
    result = runner.invoke(cli, ["startapp", "box_mod"])
    assert result.exit_code == 1
    assert "cannot start with box_" in result.output


@patch("shopyo.api.cli.get_module_path_if_exists")
def test_startapp_already_exists(mock_get_path, runner):
    mock_get_path.return_value = "/some/path"
    result = runner.invoke(cli, ["startapp", "testmod"])
    assert result.exit_code == 1
    assert "already exists" in result.output


@patch("shopyo.api.cli._collectstatic")
def test_collectstatic(mock_collect, runner):
    result = runner.invoke(cli, ["collectstatic"])
    assert result.exit_code == 0
    mock_collect.assert_called_once_with(target_module="modules", verbose=False)

    result = runner.invoke(cli, ["collectstatic", "modules/auth"])
    assert result.exit_code == 0
    mock_collect.assert_called_with(target_module="modules/auth", verbose=False)


@patch("shopyo.api.cli._clean")
def test_clean(mock_clean, runner):
    result = runner.invoke(cli, ["clean"])
    assert result.exit_code == 0
    mock_clean.assert_called_once_with(
        verbose=False, clear_migration=True, clear_db=True
    )


@patch("shopyo.api.cli._clean")
@patch("shopyo.api.cli.autoload_models")
@patch("shopyo.api.cli.run")
@patch("shopyo.api.cli._collectstatic")
@patch("shopyo.api.cli._upload_data")
def test_initialise(
    mock_upload, mock_collect, mock_run, mock_autoload, mock_clean, runner
):
    result = runner.invoke(cli, ["initialise"])
    assert result.exit_code == 0
    mock_clean.assert_called_once()
    assert mock_autoload.call_count == 2
    assert mock_run.call_count == 3  # init, migrate, upgrade
    mock_collect.assert_called_once()
    mock_upload.assert_called_once()


@patch("shopyo.api.cli._run_app")
def test_rundebug(mock_run, runner):
    result = runner.invoke(cli, ["rundebug"])
    assert result.exit_code == 0
    mock_run.assert_called_once_with("development")


@patch("shopyo.api.cli._run_app")
def test_runserver(mock_run, runner):
    result = runner.invoke(cli, ["runserver"])
    assert result.exit_code == 0
    mock_run.assert_called_once_with("production")


@patch("shopyo.api.cli._audit")
def test_audit(mock_audit, runner):
    result = runner.invoke(cli, ["audit"])
    assert result.exit_code == 0
    mock_audit.assert_called_once_with(True, True, True)

    result = runner.invoke(cli, ["audit", "--show-warning"])
    assert result.exit_code == 0
    mock_audit.assert_called_with(True, False, False)


@patch("shopyo.api.cli._rename_app")
def test_rename(mock_rename, runner):
    result = runner.invoke(cli, ["rename", "old", "new"])
    assert result.exit_code == 0
    mock_rename.assert_called_once_with("old", "new")


@patch("shopyo.api.cli.copytree")
@patch("shopyo.api.file.trymkfile")
@patch("shopyo.api.file.trymkdir")
@patch("shopyo.api.file.tryrmtree")
def test_new(mock_rmtree, mock_mkdir, mock_mkfile, mock_copytree, runner):
    # We need to patch shopyo.__version__ but since we might not invoke the real init,
    # we can try to patch where it is used or mocking the module.
    # The 'new' command imports it inside the function: from shopyo import __version__

    original_exists = os.path.exists

    def exists_side_effect(path):
        if path == "/tmp":
            return True
        if "myproj" in path or path.endswith("tmp/tmp"):
            return False
        return original_exists(path)

    with patch("shopyo.__version__", "1.0.0", create=True), patch(
        "os.getcwd", return_value="/tmp"
    ), patch("os.path.exists", side_effect=exists_side_effect):

        # Test default (current dir)
        result = runner.invoke(cli, ["new"])
        if result.exit_code != 0:
            print(result.output)
            print(result.exception)

        assert result.exit_code == 0
        assert "creating project" in result.output

        # Test with project name
        result = runner.invoke(cli, ["new", "myproj"])
        assert result.exit_code == 0
        assert "creating project myproj" in result.output


@patch("shopyo.api.cli.copytree")
def test_new_exists(mock_copytree, runner):
    with patch("shopyo.__version__", "1.0.0", create=True), patch(
        "os.getcwd", return_value="/tmp"
    ), patch("os.path.exists", return_value=True):

        result = runner.invoke(cli, ["new", "myproj"])
        assert result.exit_code == 1
        assert "Unable to create new project" in result.output



from shopyo.api import html
from markupsafe import escape


def test_notify():
    message = "Hello World"
    alert_type = "primary"

    result = html.notify(message, alert_type)

    assert message in result
    assert f"alert-{alert_type}" in result
    assert "shopyo-alert" in result
    assert "setTimeout" in result


def test_notify_xss_protection():
    # Test that dangerous characters are escaped
    dangerous_msg = "<script>alert(1)</script>"
    escaped_msg = str(escape(dangerous_msg))

    result = html.notify(dangerous_msg)

    assert dangerous_msg not in result
    assert escaped_msg in result
    assert "&lt;script&gt;alert(1)&lt;/script&gt;" in result


def test_notify_helpers():
    msg = "test message"

    # Success
    res = html.notify_success(msg)
    assert "alert-success" in res
    assert msg in res

    # Danger
    res = html.notify_danger(msg)
    assert "alert-danger" in res
    assert msg in res

    # Warning
    res = html.notify_warning(msg)
    assert "alert-warning" in res
    assert msg in res

    # Info
    res = html.notify_info(msg)
    assert "alert-info" in res
    assert msg in res



"""
Helper utility functions for commandline api
"""

import importlib
import json
import os
import re
import sys

try:
    from importlib.metadata import version
except ImportError:
    from importlib_metadata import version
from subprocess import run

import click
from flask import current_app

from shopyo.api.cli_content import get_dashboard_html_content
from shopyo.api.cli_content import get_global_py_content
from shopyo.api.cli_content import get_module_view_content
from shopyo.api.constants import SEP_CHAR
from shopyo.api.constants import SEP_NUM
from shopyo.api.file import get_folders
from shopyo.api.file import last_part_of_path
from shopyo.api.file import path_exists
from shopyo.api.file import trycopytree
from shopyo.api.file import trymkdir
from shopyo.api.file import trymkfile
from shopyo.api.file import tryrmcache
from shopyo.api.file import tryrmfile
from shopyo.api.file import tryrmtree


def _clean(verbose=False, clear_migration=True, clear_db=True):
    """
    Deletes shopyo.db and migrations/ if present in current working directory.
    Deletes all __pycache__ folders starting from current working directory
    all the way to leaf directory.

    Parameters
    ----------
        - verbose: flag to indicate whether to print to result of clean to
            stdout or not.
        - clear_migration: flag to indicate if migration folder is to be deleted or not
        - clear_db: flag indicating if db is to be cleared or not
        - db: db to be cleaned

    Returns
    -------
    None
        ...

    """
    click.echo("Cleaning...")
    click.echo(SEP_CHAR * SEP_NUM)
    try:
        db = current_app.extensions["sqlalchemy"].db
    except:
        db = current_app.extensions["sqlalchemy"]

    if clear_db:
        db.drop_all()
        sqlalchemy_version = version("sqlalchemy").split(".")

        sql = "DROP TABLE IF EXISTS alembic_version;"
        if int(sqlalchemy_version[0]) <= 1:
            db.engine.execute(sql)
        else:
            from sqlalchemy import text

            with db.engine.begin() as conn:
                result = conn.execute(text(sql))
                conn.commit()

        tryrmfile(os.path.join(os.getcwd(), "shopyo.db"), verbose=verbose)
        if verbose:
            click.echo("[x] all tables dropped")
    elif clear_db is False:
        if verbose:
            click.echo("[ ] db clearing skipped")

    tryrmcache(os.getcwd(), verbose=verbose)

    if clear_migration:
        tryrmtree(os.path.join(os.getcwd(), "migrations"), verbose=verbose)
    elif clear_migration is False:
        if verbose:
            click.echo("[ ] migration folder delete skipped")


def _collectstatic(target_module="modules", verbose=False):
    """
    Copies ``module/static`` into ``/static/modules/module``.
    In static it becomes like

    ::

       static/
            modules/
                box_something/
                    modulename
                modulename2


    Parameters
    ----------
    target_module: str
        name of module, in alphanumeric-underscore,
        supports ``module`` or ``box__name/module``

    Returns
    -------
    None

    """
    click.echo("Collecting static...")
    click.echo(SEP_CHAR * SEP_NUM)

    root_path = os.getcwd()
    static_path = os.path.join(root_path, "static")

    # if target_module path does not start with 'modules\' add it to as a
    # prefix to the target_module path
    if target_module != "modules":
        # normalize the target_module path to be same as that of OS
        target_module = re.split(r"[/|\\]+", target_module)
        target_module_start = target_module[0]
        target_module = os.path.join(*target_module)

        # add the modules folder to start of target_module incase it is not
        # already present in the path
        if target_module_start != "modules":
            target_module = os.path.join("modules", target_module)

    # get the full path for modules (the src). Defaults to ./modules
    modules_path = os.path.join(root_path, target_module)

    # get the full path of static folder to copy to (the dest).
    # always ./static/modules
    modules_path_in_static = os.path.join(static_path, "modules")

    # terminate if modules_path (i.e. src to copy static from) does not exist
    if not os.path.exists(modules_path):
        click.echo(f"[ ] path: {modules_path} does not exist")
        sys.exit(1)

    # clear ./static/modules before coping to it
    tryrmtree(modules_path_in_static, verbose=verbose)

    # look for static folders in all project
    for folder in get_folders(modules_path):
        if folder.startswith("box__"):
            box_path = os.path.join(modules_path, folder)
            for subfolder in get_folders(box_path):
                module_name = subfolder
                module_static_folder = os.path.join(box_path, subfolder, "static")
                if not os.path.exists(module_static_folder):
                    continue
                module_in_static_dir = os.path.join(
                    modules_path_in_static, folder, module_name
                )

                # copy from ./modules/<box__name>/<submodule> to
                # ./static/modules
                trycopytree(module_static_folder, module_in_static_dir, verbose=verbose)
        else:
            path_split = ""

            # split the target module if default target_module path name is
            # not used
            if target_module != "modules":
                path_split = re.split(r"[/|\\]", target_module, maxsplit=1)
                path_split = path_split[1]

            if folder.lower() == "static":
                module_static_folder = os.path.join(modules_path, folder)
                module_name = path_split
            else:
                module_static_folder = os.path.join(modules_path, folder, "static")
                module_name = os.path.join(path_split, folder)

            if not os.path.exists(module_static_folder):
                continue
            module_in_static_dir = os.path.join(modules_path_in_static, module_name)
            tryrmtree(module_in_static_dir, verbose=verbose)
            trycopytree(module_static_folder, module_in_static_dir, verbose=verbose)

    # load packages

    try:
        from init import installed_packages
    except ImportError:
        click.echo(
            "This version of Shopyo requires that\n"
            "init.py contains the line\n"
            "installed_packages = []\n"
            "please add it."
        )
        sys.exit()
    for plugin in installed_packages:
        try:
            plugin_mod = importlib.import_module(plugin)
            plugin_folder_path = plugin_mod.view.mhelp.dirpath
            plugin_static_folder = os.path.join(plugin_folder_path, "static")

            if os.path.exists(plugin_static_folder):
                plugin_in_static_dir = os.path.join(modules_path_in_static, plugin)
                tryrmtree(plugin_in_static_dir, verbose=verbose)
                trycopytree(plugin_static_folder, module_in_static_dir, verbose=verbose)
                if verbose:
                    click.echo(f"[x] collected static from {plugin}")
            else:
                if verbose:
                    click.echo(f"[ ] static folder not found: {plugin_static_folder}")
        except Exception as e:
            if verbose:
                click.echo(f"[ ] statoc {e}")

    click.echo("")


def _upload_data(verbose=False):
    click.echo("Uploading initial data to db...")
    click.echo(SEP_CHAR * SEP_NUM)

    root_path = os.getcwd()

    for folder in os.listdir(os.path.join(root_path, "modules")):
        if folder.startswith("__"):  # ignore __pycache__
            continue
        if folder.startswith("box__"):
            # boxes
            for sub_folder in os.listdir(os.path.join(root_path, "modules", folder)):
                if sub_folder.startswith("__"):  # ignore __pycache__
                    continue
                elif sub_folder.endswith(".json"):  # box_info.json
                    continue

                try:
                    upload = importlib.import_module(
                        f"modules.{folder}.{sub_folder}.upload"
                    )
                    upload.upload()
                except ImportError as e:
                    if verbose:
                        click.echo(f"[ ] {e}")
        else:
            # apps
            try:
                upload = importlib.import_module(f"modules.{folder}.upload")
                upload.upload()
            except ImportError as e:
                if verbose:
                    click.echo(f"[ ] {e}")

    # load packages

    from init import installed_packages

    for plugin in installed_packages:
        try:
            plugin_mod = importlib.import_module(f"{plugin}.upload")
            plugin_mod.upload()
            if verbose:
                click.echo(f"[x] uploaded from {plugin}")
        except Exception as e:
            if verbose:
                click.echo(f"[ ] upload info: {e}")

    click.echo("")


def _create_box(boxname, verbose=False):
    base_path = os.path.join("modules", boxname)
    trymkdir(base_path, verbose=verbose)

    info_json = {
        "display_string": boxname.capitalize(),
        "box_name": boxname,
        "author": {"name": "", "website": "", "mail": ""},
    }

    box_info = os.path.join(base_path, "box_info.json")

    with open(box_info, "w", encoding="utf-8") as f:
        json.dump(info_json, f, indent=4, sort_keys=True)

    if verbose:
        click.echo("'box_info.json' content:")
        click.echo(json.dumps(info_json, indent=4, sort_keys=True))

    click.echo(f"{boxname} created!")


def _create_module(modulename, base_path=None, verbose=False):
    """creates module with the structure defined in the modules section in docs
    Assume valid modulename i.e modulename does not start with ``box__`` and
    modulename consist only of alphanumeric characters or underscore

    Parameters
    ----------
    modulename: str
        name of module, in alphanumeric-underscore

    Returns
    -------
    None

    """

    click.echo(f"creating module: {modulename}")
    click.echo(SEP_CHAR * SEP_NUM)

    if base_path is None:
        base_path = os.path.join("modules", modulename)

    # create the module with directories templates, tests, static
    trymkdir(base_path, verbose=verbose)
    trymkdir(os.path.join(base_path, "templates"), verbose=verbose)
    trymkdir(os.path.join(base_path, "templates", modulename), verbose=verbose)
    trymkdir(os.path.join(base_path, "tests"), verbose=verbose)
    trymkdir(os.path.join(base_path, "static"), verbose=verbose)

    # create functional test and unit test files for the module
    test_func_path = os.path.join(
        base_path, "tests", f"test_{modulename}_functional.py"
    )
    test_models_path = os.path.join(base_path, "tests", f"test_{modulename}_models.py")
    test_func_content = "# Please add your functional tests to this file.\n"
    test_model_content = "# Please add your models tests to this file.\n"
    trymkfile(test_func_path, test_func_content, verbose=verbose)
    trymkfile(test_models_path, test_model_content, verbose=verbose)

    # create view.py, forms.py and model.py files inside the module
    trymkfile(
        os.path.join(base_path, "view.py"), get_module_view_content(), verbose=verbose
    )
    trymkfile(os.path.join(base_path, "forms.py"), "", verbose=verbose)
    trymkfile(os.path.join(base_path, "models.py"), "", verbose=verbose)

    # create info.json file inside the module
    info_json = {
        "display_string": modulename.capitalize(),
        "module_name": modulename,
        "type": "show",
        "fa-icon": "fa fa-store",
        "url_prefix": f"/{modulename}",
        "author": {"name": "", "website": "", "mail": ""},
    }
    info_json_path = os.path.join(base_path, "info.json")
    with open(info_json_path, "w", encoding="utf-8") as f:
        json.dump(info_json, f, indent=4, sort_keys=True)

    if verbose:
        click.echo(f"[x] file created at '{info_json_path}' with content: ")
        click.echo(json.dumps(info_json, indent=4, sort_keys=True))

    # create the sidebar.html inside templates/blocks
    blocks_path = os.path.join(base_path, "templates", modulename, "blocks")
    trymkdir(blocks_path, verbose=verbose)
    trymkfile(os.path.join(blocks_path, "sidebar.html"), "", verbose=verbose)

    # create the dashboard.html inside templates/MODULENAME
    trymkfile(
        os.path.join(base_path, "templates", modulename, "dashboard.html"),
        get_dashboard_html_content(),
        verbose=verbose,
    )

    # create the global.py files inside the module
    trymkfile(
        os.path.join(base_path, "global.py"), get_global_py_content(), verbose=verbose
    )


def _run_app(mode):
    """helper command for running shopyo flask app in debug/production mode"""
    app_path = os.path.join(os.getcwd(), "app.py")

    if not os.path.exists(app_path):
        click.secho(f"Unable to find `app.py` in {os.getcwd()}", fg="red")
        sys.exit(1)

    os.environ["FLASK_APP"] = f"app:create_app('{mode}')"
    os.environ["FLASK_ENV"] = mode
    os.environ["ENV"] = mode
    run(["flask", "run"])


def _check_modules_path(root_path):
    modules_path = os.path.join(root_path, "modules")
    if not path_exists(modules_path):
        click.echo("Modules folder not found!")
        sys.exit()


def _url_prefix_exists(url_prefix, found_url_prefixes):
    for data in found_url_prefixes:
        if url_prefix == data["url_prefix"]:
            return {"status": True, "data": data}
    return {"status": False, "data": dict()}


def _verify_app(app_path, found_url_prefixes, box_name=None):
    app_folder = last_part_of_path(app_path)

    audit_info = {"path": app_path, "issues": []}

    # verify info.json
    if not path_exists(os.path.join(app_path, "info.json")):
        audit_info["issues"].append("severe: info.json not found")
    else:
        with open(os.path.join(app_path, "info.json")) as f:
            json_data = json.load(f)

        _url_prefix_exists(json_data["url_prefix"], found_url_prefixes)

        to_check_keys = ["module_name", "url_prefix"]
        not_found = []
        for key in to_check_keys:
            if key not in json_data:
                not_found.append(key)
                msg = f"severe: key {key} not found in info.json"
                audit_info["issues"].append(msg)

        if ("module_name" not in not_found) and ("url_prefix" not in not_found):
            if (json_data["module_name"].strip() == "") or (
                json_data["url_prefix"].strip() == ""
            ):
                msg = (
                    "sus: module_name and url_prefix in info.json must not be empty"
                    " ideally"
                )
                audit_info["issues"].append(msg)

        if "module_name" not in not_found:
            if {json_data["module_name"], app_folder} != {app_folder}:
                msg = """severe: currently module_name "{}" in info.json and app folder "{}" must have the same value""".format(
                    json_data["module_name"], app_folder
                )
                audit_info["issues"].append(msg)

        url_prefix_exists = _url_prefix_exists(
            json_data["url_prefix"], found_url_prefixes
        )

        if url_prefix_exists["status"] is True:
            msg = (
                f"warning: url_prefix '{json_data['url_prefix']}' exists in"
                f" '{app_path}' and '{url_prefix_exists['data']['path']}'"
            )
            audit_info["issues"].append(msg)
        else:
            found_url_prefixes.append(
                {"path": app_path, "url_prefix": json_data["url_prefix"]}
            )

    # verify components

    if not path_exists(os.path.join(app_path, "templates")):
        audit_info["issues"].append("warning: templates folder not found")

    if not path_exists(os.path.join(app_path, "view.py")):
        audit_info["issues"].append("severe: view.py not found")

    if not path_exists(os.path.join(app_path, "forms.py")):
        audit_info["issues"].append("info: forms.py not found")

    if not path_exists(os.path.join(app_path, "models.py")):
        audit_info["issues"].append("info: models.py not found")

    if not path_exists(os.path.join(app_path, "global.py")):
        audit_info["issues"].append("info: global.py not found")

    return audit_info


def _verify_box(box_path):
    audit_info = {"issues": []}

    # verify box_info.json
    if not path_exists(os.path.join(box_path, "box_info.json")):
        audit_info["issues"].append("warning: box_info.json not found")
    else:
        with open(os.path.join(box_path, "box_info.json")) as f:
            json_data = json.load(f)

        to_check_keys = ["box_name"]

        for key in to_check_keys:
            if key not in json_data:
                msg = f"severe: key {key} not found in box_info.json"
                audit_info["issues"].append(msg)

    return audit_info


def _check_apps(root_path, found_url_prefixes):
    issues_found = []

    modules_path = os.path.join(root_path, "modules")
    apps = get_folders(modules_path)
    apps = [a for a in apps if not a.startswith("__") and not a.startswith("box__")]

    for app in apps:
        app_path = os.path.join(modules_path, app)
        app_issues = _verify_app(app_path, found_url_prefixes)
        issues_found.append(app_issues)

    return issues_found


def _check_boxes(root_path, found_url_prefixes):
    box_issues = []

    modules_path = os.path.join(root_path, "modules")
    boxes = get_folders(modules_path)
    boxes = [b for b in boxes if b.startswith("box__")]

    for b in boxes:
        box_info = {"path": os.path.join(modules_path, b), "apps_issues": []}
        box_info["issues"] = _verify_box(os.path.join(modules_path, b))["issues"]

        for app in get_folders(os.path.join(modules_path, b)):
            app_issues = _verify_app(
                os.path.join(modules_path, b, app), found_url_prefixes, box_name=b
            )
            box_info["apps_issues"].append(app_issues)
        box_issues.append(box_info)

    return box_issues


def _audit(warning, info, severe):
    """
    checks if modules are corrupted
    """
    issue_type = {"warning": warning, "info": info, "severe": severe}
    found_url_prefixes = []

    root_path = os.getcwd()
    _check_modules_path(root_path)
    apps_issues = _check_apps(root_path, found_url_prefixes)
    boxes_issues = _check_boxes(root_path, found_url_prefixes)

    click.echo("Running audit ...")

    click.echo("Checking apps ...")
    for app_issue in apps_issues:
        click.echo(app_issue["path"])
        for issue in app_issue["issues"]:
            if issue_type[issue.split(":")[0]]:
                click.echo("    " + issue)
        click.echo("")

    for box_issue in boxes_issues:
        click.echo(box_issue["path"])
        for issue in box_issue["issues"]:
            if issue_type[issue.split(":")[0]]:
                click.echo("    " + issue)

        for app_issue in box_issue["apps_issues"]:
            click.echo("    " + app_issue["path"])
            for issue in app_issue["issues"]:
                if issue_type[issue.split(":")[0]]:
                    click.echo("        " + issue)
        click.echo("")

    click.echo("Audit finished!")


def _verify_app_name(app_name):
    if (app_name.startswith("box__")) and (app_name.count("/") != 1):
        return False

    if app_name.endswith("/"):
        return False

    return True


def name_is_box(app_name):
    if app_name.startswith("box__"):
        return True

    return False


def _rename_app(old_app_name, new_app_name):
    # box_
    if old_app_name.startswith("box") and not old_app_name.startswith("box__"):
        click.echo('Box names start with two __, example: "box__default"')
        sys.exit()
    if new_app_name.startswith("box") and not new_app_name.startswith("box__"):
        click.echo('Box names start with two __, example: "box__default"')
        sys.exit()

    if (not _verify_app_name(old_app_name)) and (not _verify_app_name(new_app_name)):
        click.echo('App names should be in the format "app" or "box_name/app"')
        sys.exit()

    root_path = os.getcwd()
    modules_path = os.path.join(root_path, "modules")

    if name_is_box(old_app_name):
        box_name = old_app_name.split("/")[0]
        if name_is_box(new_app_name):
            app_part = old_app_name.split("/")[1]
            new_app_name.split("/")[1]

        if not path_exists(os.path.join(modules_path, box_name, app_part)):
            click.echo(f"App {old_app_name} does not exist")
            sys.exit()

    try:
        os.rename(
            os.path.join(modules_path, old_app_name),
            os.path.join(modules_path, new_app_name),
        )

        with open(os.path.join(modules_path, new_app_name, "info.json")) as f:
            json_data = json.load(f)

        with open(os.path.join(modules_path, new_app_name, "info.json"), "w+") as f:
            if name_is_box(new_app_name):
                module_name = new_app_name.split("/")[1]
            else:
                module_name = new_app_name
            json_data["module_name"] = module_name
            json.dump(json_data, f, indent=4)

        click.echo(f"Renamed app {old_app_name} to {new_app_name}")
    except Exception as e:
        raise e



import os
import pytest
import sys
from unittest.mock import patch, MagicMock
from shopyo.api.cmd import (
    clean,
    create_module,
    create_box,
    createmodulebox,
    initialise,
    collect_static,
    create_module_in_box,
)


@pytest.fixture
def mock_app():
    app = MagicMock()
    return app


def test_clean_cmd(mock_app):
    # Mocking 'from init import db'
    mock_db = MagicMock()
    with patch("shopyo.api.cmd.db", mock_db), patch(
        "shopyo.api.cmd.tryrmcache"
    ) as mock_rmcache, patch("shopyo.api.cmd.tryrmfile") as mock_rmfile, patch(
        "shopyo.api.cmd.tryrmtree"
    ) as mock_rmtree:

        # We also need to mock the context manager
        mock_app.test_request_context.return_value.__enter__.return_value = MagicMock()

        clean(mock_app)
        mock_db.drop_all.assert_called_once()
        mock_rmcache.assert_called_once()
        mock_rmfile.assert_called_once()
        mock_rmtree.assert_called_once()


@patch("shopyo.api.cmd.trymkdir")
@patch("shopyo.api.cmd.trymkfile")
def test_create_module_cmd(mock_mkfile, mock_mkdir):
    create_module("testmod")
    mock_mkdir.assert_called()
    mock_mkfile.assert_called()


@patch("shopyo.api.cmd.trymkdir")
@patch("shopyo.api.cmd.trymkfile")
def test_create_box_cmd(mock_mkfile, mock_mkdir):
    with patch("shopyo.api.cmd.os.path.exists", return_value=False):
        create_box("testbox")
        mock_mkdir.assert_called_once()
        mock_mkfile.assert_called_once()


@patch("shopyo.api.cmd.create_module")
@patch("shopyo.api.cmd.create_box")
def test_createmodulebox(mock_create_box, mock_create_module):
    # Test module
    res = createmodulebox("testmod")
    assert res[0] is True
    mock_create_module.assert_called_with("testmod")

    # Test box
    res = createmodulebox("box__test")
    assert res[0] is True
    mock_create_box.assert_called_with("box__test")

    # Test module in box
    with patch("shopyo.api.cmd.create_module_in_box") as mock_mod_in_box:
        res = createmodulebox("box__test/mod")
        assert res[0] is True
        mock_mod_in_box.assert_called_with("mod", "box__test")


@patch("shopyo.api.cmd.os.path.exists", return_value=True)
@patch("shopyo.api.cmd.tryrmtree")
@patch("shopyo.api.cmd.get_folders", return_value=[])
def test_collect_static(mock_get_folders, mock_rmtree, mock_exists):
    collect_static()
    mock_rmtree.assert_called()


@patch("shopyo.api.cmd.os.path.exists")
@patch("shopyo.api.cmd.create_module")
def test_create_module_in_box(mock_create_module, mock_exists):
    # Success case
    mock_exists.side_effect = [True, False]  # box exists, module doesn't
    create_module_in_box("mod", "box__test")
    mock_create_module.assert_called_once()

    # Box doesn't exist
    mock_exists.side_effect = [False]
    with patch("shopyo.api.cmd.os.listdir", return_value=[]):
        create_module_in_box("mod", "box__test")
        # Should print and not call create_module
        assert mock_create_module.call_count == 1



import os

from flask_login import LoginManager
from flask_mailman import Mail
from flask_migrate import Migrate
from flask_sqlalchemy import SQLAlchemy
from flask_wtf.csrf import CSRFProtect


root_path = os.path.dirname(os.path.abspath(__file__))  # don't remove
static_path = os.path.join(root_path, "static")  # don't remove
modules_path = os.path.join(root_path, "modules")  # don't remove
themes_path = os.path.join(static_path, "themes")  # don't remove
installed_packages = []  # don't remove


db = SQLAlchemy()
login_manager = LoginManager()
migrate = Migrate()
mail = Mail()
csrf = CSRFProtect()
login_manager = LoginManager()


import importlib


def load_extensions(app):
    migrate.init_app(app, db)
    db.init_app(app)
    mail.init_app(app)
    login_manager.init_app(app)
    csrf.init_app(app)

    with app.app_context():
        for plugin in app.extensions:
            if plugin.startswith("shopyo_"):
                try:
                    module = importlib.import_module(f"{plugin}.models")
                    for attr_name in dir(module):
                        attr = getattr(module, attr_name)
                except Exception as e:
                    print(e)



import os
import shutil
import uuid

import click

# from werkzeug.utils import secure_filename


def tryrmcache(dir_name, verbose=False):
    """
    removes all __pycache__ starting from directory dir_name
    all the way to leaf directory

    Args:
        dir_name(string) : path from where to start removing pycache
    """
    # directory_list = list()
    is_removed = False
    for root, dirs, _ in os.walk(dir_name, topdown=False):
        for name in dirs:
            # directory_list.append(os.path.join(root, name))
            if name == "__pycache__":
                shutil.rmtree(os.path.join(root, name))
                is_removed = True

    if verbose:
        if is_removed:
            click.echo("[x] __pycache__ successfully deleted")
        else:
            click.echo("[ ] __pycache__ doesn't exist", err=True)

    return is_removed


def tryrmfile(path, verbose=False):
    """
    tries to remove file path and output message to stdin or stderr.
    Path must point to a file

    Args:
        path (string): path of the file to remove

    Returns:
        bool: returns true upon successful removal false otherwise
    """
    try:
        os.remove(path)
        if verbose:
            click.echo(f"[x] file '{path}' successfully deleted")
        return True
    except OSError as e:
        if verbose:
            click.echo(
                f"[ ] unable to delete {e.filename}: {e.strerror}",
                err=True,
            )
        return False


def tryrmtree(path, verbose=False):
    """
    Tries to removes an entire directory tree. Path must point to
    a directory. Outputs message to stdin or stderr

    Args:
        path (string): directory path to be removed

    Returns:
        bool: returns true upon successful return false otherwise
    """
    try:
        shutil.rmtree(path)
        if verbose:
            click.echo(f"[x] folder '{path}' successfully deleted")
        return True
    except OSError as e:
        if verbose:
            click.echo(
                f"[ ] unable to delete {e.filename}: {e.strerror}",
                err=True,
            )
        return False


def trycopytree(source, dest, verbose=False):
    """
    Recursive copy of folder

    Parameters
    ----------
    source: str
        source folder path
    dest: str
        destination folder path

    Returns
    -------
    None
    """
    try:
        shutil.copytree(source, dest)
        if verbose:
            click.echo(f"[x] done copying {source} to {dest}")
    except Exception as e:
        click.echo(f"[ ] unable to copy directory tree. {e}", err=True)


def trycopy(source, dest, verbose=False):
    """
    Non-ecursive copy of folder

    Parameters
    ----------
    source: str
        source folder path
    dest: str
        destination folder path

    Returns
    -------
    None
    """
    try:
        shutil.copy(source, dest)
        if verbose:
            click.echo(f"[x] done copying {source} to {dest}")
    except Exception as e:
        click.echo(f"[ ] unable to copy file. {e}", err=True)


def trymkdir(path, verbose=False):
    """
    Creates dir at destination

    Parameters
    ----------
    path: str
        path with folder already in

    Returns
    -------
    None
    """
    try:
        os.mkdir(path)
        if verbose:
            click.echo(f"[x] Successfully created dir {path}")
    except Exception as e:
        click.echo(f"[ ] unable to make directory. {e}", err=True)


def trymkfile(path, content, verbose=False):
    """
    Creates file

    Parameters
    ----------
    path: str
        path to create file with filename included
    content: str
        file content

    Returns
    -------
    None
    """
    try:
        with open(path, "w+") as f:
            f.write(content)
        if verbose:
            click.echo(f"[x] file {path} created with content: ")
            click.echo(content)
    except Exception as e:
        click.echo(f"[ ] {e}")


def absdiroffile(filepath):
    """
    Gives absolute directory of file, normally expects __file__ as
    param

    Parameters
    ----------
    filepath: str
        path of file

    Returns
    -------
    str
        Absolute dir path of file
    """
    return os.path.dirname(os.path.abspath(filepath))


def get_folders(path):
    dirs = [d for d in os.listdir(path) if os.path.isdir(os.path.join(path, d))]
    return dirs


def unique_filename(fname):
    prepended = str(uuid.uuid4()).replace("-", "_")[:10]
    return f"{prepended}_{fname}"


def path_exists(path):
    return os.path.exists(path)


def last_part_of_path(path):
    return os.path.basename(os.path.normpath(path))


def delete_file(path):
    os.remove(path)


# def unique_sec_filename(filename):
#     return unique_filename(secure_filename(filename))



"""
Enhanced DB-related helper utilities for Flask applications.
Provides powerful mixins and base classes for rapid development.
"""

import os
import sys
from datetime import datetime, timezone
from typing import Any, Dict, List, Optional, Type, TypeVar, Union
from sqlalchemy import Column, DateTime, Boolean, func
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import Query

sys.path.append(os.getcwd())
from init import db

T = TypeVar("T", bound="YoModel")

# Use timezone.utc for compatibility with Python < 3.11
UTC = timezone.utc


class TimestampMixin:
    """Mixin that adds created_at and updated_at timestamps."""

    created_at = Column(DateTime, default=lambda: datetime.now(UTC), nullable=False)
    updated_at = Column(
        DateTime,
        default=lambda: datetime.now(UTC),
        onupdate=lambda: datetime.now(UTC),
        nullable=False,
    )


class SoftDeleteMixin:
    """Mixin that adds soft delete functionality."""

    is_deleted = Column(Boolean, default=False, nullable=False)

    def soft_delete(self, commit: bool = True) -> Optional["SoftDeleteMixin"]:
        """Soft delete the record by setting is_deleted to True."""
        self.is_deleted = True
        if commit:
            self.save()
            return self
        return None

    @classmethod
    def get_active(cls) -> Query:
        """Get all non-deleted records."""
        return cls.query.filter_by(is_deleted=False)


class ValidationMixin:
    """Mixin that adds validation functionality."""

    def validate(self) -> List[str]:
        """Validate the model instance.

        Returns:
            List[str]: List of validation errors, empty if valid
        """
        return []

    def is_valid(self) -> bool:
        """Check if the model instance is valid."""
        return len(self.validate()) == 0


class SearchMixin:
    """Mixin that adds search functionality."""

    @classmethod
    def search(cls, query: str, fields: List[str]) -> Query:
        """Search records by query string in specified fields.

        Args:
            query (str): Search query
            fields (List[str]): List of field names to search in

        Returns:
            Query: SQLAlchemy query with search filters
        """
        search_conditions = []
        for field in fields:
            if hasattr(cls, field):
                search_conditions.append(getattr(cls, field).ilike(f"%{query}%"))
        return cls.query.filter(db.or_(*search_conditions))


class PaginationMixin:
    """Mixin that adds pagination functionality."""

    @classmethod
    def paginate(cls, page: int = 1, per_page: int = 20) -> Dict[str, Any]:
        """Paginate records.

        Args:
            page (int): Page number
            per_page (int): Number of items per page

        Returns:
            Dict[str, Any]: Dictionary containing paginated items and metadata
        """
        pagination = cls.query.paginate(page=page, per_page=per_page)
        return {
            "items": pagination.items,
            "total": pagination.total,
            "pages": pagination.pages,
            "current_page": pagination.page,
            "has_next": pagination.has_next,
            "has_prev": pagination.has_prev,
        }


class CRUDMixin:
    """
    Mixin that adds convenience methods for
    CRUD (create, read, update, delete) operations.
    """

    @classmethod
    def create(cls: Type[T], **kwargs) -> T:
        """Create a new record and save it in the database.

        Returns:
            T: The created record
        """
        instance = cls(**kwargs)
        return instance.save()

    def update(self, commit: bool = True, **kwargs) -> Optional[T]:
        """Update specific fields of a record.

        Args:
            commit (bool): Whether to commit the changes
            **kwargs: Fields to update

        Returns:
            Optional[T]: The updated record if committed, None otherwise
        """
        for attr, value in kwargs.items():
            setattr(self, attr, value)
        if commit:
            self.save()
            return self
        return None

    def save(self, commit: bool = True) -> T:
        """Save the record.

        Args:
            commit (bool): Whether to commit the changes

        Returns:
            T: The saved record
        """
        db.session.add(self)
        if commit:
            db.session.commit()
        return self

    def delete(self, commit: bool = True) -> Optional[T]:
        """Remove the record from the database.

        Args:
            commit (bool): Whether to commit the changes

        Returns:
            Optional[T]: The deleted record if committed, None otherwise
        """
        db.session.delete(self)
        if commit:
            db.session.commit()
            return self
        return None

    @classmethod
    def bulk_create(cls: Type[T], items: List[Dict[str, Any]]) -> List[T]:
        """Create multiple records at once.

        Args:
            items (List[Dict[str, Any]]): List of dictionaries containing record data

        Returns:
            List[T]: List of created records
        """
        instances = [cls(**item) for item in items]
        db.session.bulk_save_objects(instances)
        db.session.commit()
        return instances


class YoModel(CRUDMixin, db.Model):
    """Base model class that includes CRUD convenience methods."""

    __abstract__ = True

    @declared_attr
    def __tablename__(cls) -> str:
        """Generate __tablename__ automatically."""
        return cls.__name__.lower()


class PkModel(YoModel):
    """
    Base model class that includes CRUD convenience methods,
    plus adds a 'primary key' column named 'id'.
    """

    __abstract__ = True
    id = db.Column(db.Integer, primary_key=True)

    @classmethod
    def get_by_id(cls: Type[T], record_id: Union[int, str]) -> Optional[T]:
        """Get record by ID.

        Args:
            record_id (Union[int, str]): ID of record to get

        Returns:
            Optional[T]: Object identified by record_id if any, None otherwise
        """
        if any(
            (
                isinstance(record_id, (str, bytes)) and record_id.isdigit(),
                isinstance(record_id, (int, float)),
            )
        ):
            return cls.query.get(int(record_id))
        return None

    @classmethod
    def get_or_404(cls: Type[T], record_id: Union[int, str]) -> T:
        """Get record by ID or raise 404 error.

        Args:
            record_id (Union[int, str]): ID of record to get

        Returns:
            T: Object identified by record_id

        Raises:
            werkzeug.exceptions.NotFound: If record not found
        """
        from flask import abort

        rv = cls.get_by_id(record_id)
        if rv is None:
            abort(404)
        return rv



from shopyo.api.module import ModuleHelp

# from flask import render_template
# from flask import url_for
# from flask import redirect
# from flask import flash
# from flask import request

# from shopyo.api.html import notify_success
# from shopyo.api.forms import flash_errors

mhelp = ModuleHelp(__file__, __name__)
globals()[mhelp.blueprint_str] = mhelp.blueprint
module_blueprint = globals()[mhelp.blueprint_str]


@module_blueprint.route("/")
def index():
    return mhelp.info["display_string"]


# If "dashboard": "/dashboard" is set in info.json
#
# @module_blueprint.route("/dashboard", methods=["GET"])
# def dashboard():

#     context = mhelp.context()

#     context.update({

#         })
#     return mhelp.render('dashboard.html', **context)



import re
import subprocess
import sys
from pathlib import Path

from shopyo.api.info import printinfo

dirpath = Path(__file__).parent.absolute()
dirpathparent = Path(__file__).parent.parent.absolute()


def is_venv():
    return hasattr(sys, "real_prefix") or (
        hasattr(sys, "base_prefix") and sys.base_prefix != sys.prefix
    )


def is_valid_name(name):
    notallowedpattern = r"[_\.]+"
    allowedpattern = r"^[\w+\.]+$"
    isallowed = re.match(allowedpattern, name)
    isnotallowed = re.match(notallowedpattern, name)

    if not isnotallowed and isallowed:
        return True
    else:
        return False


def main():
    args = sys.argv
    if len(args) == 1:
        printinfo()
        print("No arguments supplied")
    else:
        if not is_venv():
            print("Please use Shopyo in a virtual environment for this command")
            sys.exit()
        torun = [sys.executable, "manage.py"] + args[1:]
        subprocess.run(torun)


if __name__ == "__main__":
    main()
